@@ -1,7 +1,7 @@
 /*
  * AIDE (Advanced Intrusion Detection Environment)
  *
- * Copyright (C) 1999-2006, 2010-2011, 2013, 2015-2016, 2019-2021 Rami Lehti,
+ * Copyright (C) 1999-2006, 2010-2011, 2013, 2015-2016, 2019-2023 Rami Lehti,
  *               Pablo Virolainen, Richard van den Berg, Hannes von Haugwitz
  *
  * This program is free software; you can redistribute it and/or
@@ -24,9 +24,16 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <errno.h>
-#include <sys/param.h>
-
+#include <stdbool.h>
+#include <zlib.h>
+#include "attributes.h"
+#include "conf_ast.h"
+#include "config.h"
+#include "hashsum.h"
+#include "list.h"
+#include "rx_rule.h"
+#include "seltree_struct.h"
+#include "url.h"
 #include "commandconf.h"
 #include "conf_lex.h"
 #include "log.h"
@@ -34,18 +41,25 @@
 #include "db.h"
 #include "db_config.h"
 #include "report.h"
-#include "gen_list.h"
 #include "symboltable.h"
 #include "md.h"
 #include "util.h"
-#include "base64.h"
 #include "conf_eval.h"
+#include "seltree.h"
 /*for locale support*/
 #include "locale-aide.h"
 /*for locale support*/
 #ifdef WITH_CURL
 #include "fopen.h"
 #endif
+#ifdef WITH_E2FSATTRS
+#include "e2fsattrs.h"
+#endif
+#ifdef WITH_PTHREAD
+#include <math.h>
+#include <errno.h>
+#include <unistd.h>
+#endif
 
 #define BUFSIZE 4096
 #define ZBUFSIZE 16384
@@ -81,13 +95,13 @@ url_t* parse_url(char* val, int linenumber, char* filename, char* linebuf)
       r+=2;
       for(i=0;r[0]!='/'&&r[0]!='\0';r++,i++);
       if(r[0]=='\0'){
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, invalid file-URL '%s': no path after hostname, val)
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "invalid file-URL '%s': no path after hostname", val)
     free(val_copy);
     free(u);
     return NULL;
       }
       if( (strcmp(t,"localhost") != 0) && !( conf->hostname && strcmp(t,conf->hostname)==0)){
-          LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, invalid file-URL '%s': cannot use hostname other than 'localhost' or '%s', val, conf->hostname);
+          LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "invalid file-URL '%s': cannot use hostname other than 'localhost' or '%s'", val, conf->hostname);
           free(u);
           free(val_copy);
           return NULL;
@@ -106,7 +120,7 @@ url_t* parse_url(char* val, int linenumber, char* filename, char* linebuf)
 #ifdef WITH_CURL
     u->value=checked_strdup(val);
 #else
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, %s, "http, https and ftp URL support not compiled in, recompile AIDE with '--with-curl'")
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "%s", "http, https and ftp URL support not compiled in, recompile AIDE with '--with-curl'")
     free(val_copy);
     free(u);
     return NULL;
@@ -124,7 +138,7 @@ url_t* parse_url(char* val, int linenumber, char* filename, char* linebuf)
 #ifdef HAVE_SYSLOG
     u->value=checked_strdup(r);
 #else
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, %s, "syslog url support not compiled in, recompile AIDE with syslog support")
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "%s", "syslog url support not compiled in, recompile AIDE with syslog support")
     free(val_copy);
     free(u);
     return NULL;
@@ -133,7 +147,8 @@ url_t* parse_url(char* val, int linenumber, char* filename, char* linebuf)
   }
   }
   } else {
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, unknown URL-type: '%s', val_copy);
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "unknown URL-type: '%s'", val_copy);
+    free(val_copy);
     free(u);
     return NULL;
   }
@@ -156,7 +171,7 @@ int parse_config(char *before, char *config, char* after) {
           return RETFAIL;
         }
         conf_lex_delete_buffer();
-        eval_config(config_ast, 0);
+        eval_config(config_ast, 0, NULL);
         deep_free(config_ast);
         config_ast = NULL;
     }
@@ -166,7 +181,7 @@ int parse_config(char *before, char *config, char* after) {
           return RETFAIL;
         }
         conf_lex_delete_buffer();
-        eval_config(config_ast, 0);
+        eval_config(config_ast, 0, NULL);
         deep_free(config_ast);
         config_ast = NULL;
     }
@@ -176,7 +191,7 @@ int parse_config(char *before, char *config, char* after) {
           return RETFAIL;
         }
         conf_lex_delete_buffer();
-        eval_config(config_ast, 0);
+        eval_config(config_ast, 0, NULL);
         deep_free(config_ast);
         config_ast = NULL;
     }
@@ -198,7 +213,7 @@ int db_input_wrapper(char* buf, int max_size, database* db)
   log_msg(LOG_LEVEL_TRACE,"db_input_wrapper(): parameters: buf=%p, max_size=%d, db=%p)", buf, max_size, db);
   int retval=0;
 #ifdef WITH_ZLIB
-  int c=0;
+  int c;
 #endif
 
 #ifdef WITH_CURL
@@ -226,8 +241,8 @@ int db_input_wrapper(char* buf, int max_size, database* db)
       c=fgetc(db->fp);
       if(c==(unsigned char)'\213'){
     log_msg(LOG_LEVEL_DEBUG,"db_input_wrapper(): handle gzip header");
-    lseek(fileno(db->fp),0L,SEEK_SET);
-    db->gzp=gzdopen(fileno(db->fp),"rb");
+    lseek(fileno((FILE *)db->fp),0L,SEEK_SET);
+    db->gzp=gzdopen(fileno((FILE *)db->fp),"rb");
     c=gzgetc(db->gzp);
     log_msg(LOG_LEVEL_DEBUG, "db_input_wrapper(): first character after gzip header is: %c(%#X)\n",c,c);
   if(c==-1) {
@@ -265,14 +280,14 @@ void do_define(char* name, char* value, int linenumber, char* filename, char* li
   list* l=NULL;
 
   if(!(l=list_find(name,conf->defsyms))){
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, define '%s' with value '%s', name, value)
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "define '%s' with value '%s'", name, value)
     s=(symba*)checked_malloc(sizeof(symba));
     s->name=checked_strdup(name);
     s->value=value;
     conf->defsyms=list_append(conf->defsyms,(void*)s);
   }
   else {
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, redefine '%s' with value '%s' (previous value: '%s'), name, value, ((symba*)l->data)->value)
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "redefine '%s' with value '%s' (previous value: '%s')", name, value, ((symba*)l->data)->value)
     free(((symba*)l->data)->value);
     ((symba*)l->data)->value=NULL;
     ((symba*)l->data)->value=value;
@@ -284,18 +299,18 @@ void do_undefine(char* name, int linenumber, char* filename, char* linebuf)
   list*r=NULL;
 
   if((r=list_find(name,conf->defsyms))){
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, undefine '%s' (value: '%s'), name, ((symba*)r->data)->value)
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "undefine '%s' (value: '%s')", name, ((symba*)r->data)->value)
     free(((symba*)r->data)->name);
     free(((symba*)r->data)->value);
     free((symba*)r->data);
     r->data=NULL;
     conf->defsyms=list_delete_item(r);
   } else {
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, variable '%s' to be undefined not found, name);
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "variable '%s' to be undefined not found", name);
   }
 }
 
-bool add_rx_rule_to_tree(char* rx, RESTRICTION_TYPE restriction, DB_ATTR_TYPE attr, int type, seltree *tree, int linenumber, char* filename, char* linebuf) {
+bool add_rx_rule_to_tree(char* rx, char* rule_prefix, RESTRICTION_TYPE restriction, DB_ATTR_TYPE attr, int type, seltree *tree, int linenumber, char* filename, char* linebuf) {
 
     rx_rule* r=NULL;
 
@@ -304,29 +319,43 @@ bool add_rx_rule_to_tree(char* rx, RESTRICTION_TYPE restriction, DB_ATTR_TYPE at
     char *attr_str = NULL;
     char *rs_str = NULL;
 
-    const char* rule_error;
-    int         rule_erroffset;
+    char *regex;
+    if (rule_prefix) {
+        int length = strlen(rule_prefix)+strlen(rx);
+        regex = checked_malloc(length+1);
+        strncpy(regex, rule_prefix, length+1);
+        strncat(regex, rx, length-strlen(rule_prefix)+1);
+        log_msg(LOG_LEVEL_DEBUG, "prepend regex '%s' with prefix '%s': '%s'", rx, rule_prefix, regex);
+    } else {
+        regex = checked_strdup(rx);
+    }
 
-    if ((r = add_rx_to_tree(rx, restriction, type, tree, &rule_error, &rule_erroffset)) == NULL) {
-        log_msg(LOG_LEVEL_ERROR, "%s:%d:%i: error in rule '%s': %s (line: '%s')", filename, linenumber, rule_erroffset, rx, rule_error, linebuf);
+    if ((r = add_rx_to_tree(regex, restriction, type, tree, linenumber, filename, linebuf)) == NULL) {
         retval = false;
     }else {
         r->config_linenumber = linenumber;
         r->config_filename = filename;
         r->config_line = checked_strdup(linebuf);
+        r->prefix = rule_prefix;
 
         DB_ATTR_TYPE unsupported_hashes = attr&(get_hashes(true)&~get_hashes(false));
         if (unsupported_hashes) {
             char *str;
-            LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_WARNING, ignoring unsupported hash algorithm(s): %s, str = diff_attributes(0, unsupported_hashes));
+            LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_WARNING, "ignoring unsupported hash algorithm(s): %s", str = diff_attributes(0, unsupported_hashes));
             free(str);
             attr &= ~unsupported_hashes;
         }
 
         r->attr=attr;
+        if (attr&ATTR(attr_sizeg)) {
+            log_msg(LOG_LEVEL_NOTICE, "%s:%d: Using 'S' attribute is DEPRECATED and will be removed in the release after next. Update your config and use 'growing+s' instead (line: '%s')", filename, linenumber, linebuf);
+        }
+        if (attr&ATTR(attr_compressed) && !(attr&get_hashes(false))) {
+            log_msg(LOG_LEVEL_WARNING, "%s:%d: ignore 'comprressed' attribute (no hashsum attributes are set) (line: '%s')", filename, linenumber, linebuf);
+        }
         conf->db_out_attrs |= attr;
 
-        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, add %s '%s%s %s %s' to node '%s', get_rule_type_long_string(type), get_rule_type_char(type), r->rx, rs_str = get_restriction_string(r->restriction), attr_str = diff_attributes(0, r->attr),  (r->node)->path)
+        LOG_CONFIG_FORMAT_LINE_PREFIX(LOG_LEVEL_CONFIG, "add %s '%s%s %s %s' to node '%s'", get_rule_type_long_string(type), get_rule_type_char(type), r->rx, rs_str = get_restriction_string(r->restriction), attr_str = diff_attributes(0, r->attr),  (r->node)->path)
         free(rs_str);
         free(attr_str);
 
@@ -399,7 +428,7 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
           case url_stdout:
           case url_stderr:
           case url_syslog: {
-              LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, '%s': unsupported URL-type: '%s', db_option_name, get_url_type_string(u->type))
+              LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "'%s': unsupported URL-type: '%s'", db_option_name, get_url_type_string(u->type))
               return false;
           }
           case url_stdin:
@@ -408,6 +437,7 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
           case url_https:
           case url_fd:
           case url_file:
+                /* url type is supported */
                 break;
         }
         break;
@@ -417,7 +447,7 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
           case url_stdin:
           case url_stderr:
           case url_syslog: {
-              LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, '%s': unsupported URL-type: '%s', db_option_name, get_url_type_string(u->type))
+              LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "'%s': unsupported URL-type: '%s'", db_option_name, get_url_type_string(u->type))
               return false;
           }
           case url_stdout:
@@ -426,6 +456,7 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
           case url_https:
           case url_fd:
           case url_file:
+                /* url type is supported */
                 break;
         }
         break;
@@ -435,12 +466,12 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
     db->linenumber = linenumber;
     db->filename = filename;
     db->linebuf = linebuf?checked_strdup(linebuf):NULL;
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, set '%s' option to '%s:%s', db_option_name, get_url_type_string(u->type), u->value)
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "set '%s' option to '%s:%s'", db_option_name, get_url_type_string(u->type), u->value)
     } else {
         return false;
     }
   } else {
-    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, '%s' option already set to '%s:%s' (ignore new value '%s'), db_option_name, get_url_type_string((db->url)->type), (db->url)->value, val);
+    LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "'%s' option already set to '%s:%s' (ignore new value '%s')", db_option_name, get_url_type_string((db->url)->type), (db->url)->value, val);
   }
   return true;
 }
@@ -448,7 +479,7 @@ bool do_dbdef(DB_TYPE dbtype ,char* val, int linenumber, char* filename, char* l
 bool do_repurldef(char* val, int linenumber, char* filename, char* linebuf) {
     url_t* u = parse_url(val, linenumber, filename, linebuf);
     if (add_report_url(u, linenumber, filename, linebuf)) {
-        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, set 'report_url' to '%s%s%s', get_url_type_string(u->type), u->value?":":"", u->value?u->value:"")
+        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "set 'report_url' to '%s%s%s'", get_url_type_string(u->type), u->value?":":"", u->value?u->value:"")
             return true;
     }
     return false;
@@ -460,83 +491,61 @@ bool do_reportlevel(char* val, int linenumber, char* filename, char* linebuf) {
   report_level = get_report_level(val);
   if (report_level) {
       conf->report_level = report_level;
-      LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, set 'report_level' option to '%s' (raw: %d), val, report_level)
+      LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "set 'report_level' option to '%s' (raw: %d)", val, report_level)
       return true;
   } else {
-      LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, invalid report level: '%s', val);
+      LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_ERROR, "invalid report level: '%s'", val);
       return false;
   }
 }
 
-void do_rootprefix(char* val, int linenumber, char* filename, char* linebuf) {
+bool do_rootprefix(char* val, int linenumber, char* filename, char* linebuf) {
     if (conf->root_prefix == NULL) {
         conf->root_prefix=val;
         conf->root_prefix_length=strlen(conf->root_prefix);
         if (conf->root_prefix_length && conf->root_prefix[conf->root_prefix_length-1] == '/') {
             conf->root_prefix[--conf->root_prefix_length] = '\0';
-            log_msg(LOG_LEVEL_NOTICE, "%s:%d: removed trailing '/' from root prefix", filename, linenumber);
+            LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "removed trailing '/' from root prefix: '%s'", conf->root_prefix);
         }
-        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, set 'root_prefix' option to '%s', conf->root_prefix)
+        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_CONFIG, "set 'root_prefix' option to '%s'", conf->root_prefix)
+        return true;
     } else {
-        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, 'root_prefix' option already set to '%s' (ignore new value '%s'), conf->root_prefix, val);
-        free(val);
+        LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "'root_prefix' option already set to '%s' (ignore new value '%s')", conf->root_prefix, val);
+        return false;
     }
 }
 
+#ifdef WITH_PTHREAD
+long do_num_workers(const char *str) {
+    char *err;
+    long number = strtol(str,&err,10);
+    if (err[0] == '%' && err[1] == '\0') {
+        if (number >= 0 && number <= 100) {
+            long num_of_processors = sysconf(_SC_NPROCESSORS_ONLN);
+            return ceill(num_of_processors * number / 100.0);
+        } else {
+            return -1;
+        }
+    } else if(*err != '\0' || number < 0 || errno == ERANGE) {
+        return -1;
+    }
+    return number;
+}
+#endif
+
 #ifdef WITH_E2FSATTRS
-#define easy_e2fsattrs_case(c,f) \
-case c: { \
-    conf->report_ignore_e2fsattrs|=f; \
-    break; \
-}
-
 void do_report_ignore_e2fsattrs(char* val, int linenumber, char* filename, char* linebuf) {
     conf->report_ignore_e2fsattrs = 0UL;
-    while (*val) {
-        switch(*val){
-            /* source for mappings see report.c */
-            easy_e2fsattrs_case('s',EXT2_SECRM_FL)
-            easy_e2fsattrs_case('u',EXT2_UNRM_FL)
-            easy_e2fsattrs_case('S',EXT2_SYNC_FL)
-            easy_e2fsattrs_case('D',EXT2_DIRSYNC_FL)
-            easy_e2fsattrs_case('i',EXT2_IMMUTABLE_FL)
-            easy_e2fsattrs_case('a',EXT2_APPEND_FL)
-            easy_e2fsattrs_case('d',EXT2_NODUMP_FL)
-            easy_e2fsattrs_case('A',EXT2_NOATIME_FL)
-            easy_e2fsattrs_case('c',EXT2_COMPR_FL)
-            easy_e2fsattrs_case('B',EXT2_COMPRBLK_FL)
-            easy_e2fsattrs_case('Z',EXT2_DIRTY_FL)
-            easy_e2fsattrs_case('X',EXT2_NOCOMPR_FL)
-#ifdef EXT2_ECOMPR_FL
-            easy_e2fsattrs_case('E',EXT2_ECOMPR_FL)
-#else
-            easy_e2fsattrs_case('E',EXT4_ENCRYPT_FL)
-#endif
-            easy_e2fsattrs_case('j',EXT3_JOURNAL_DATA_FL)
-            easy_e2fsattrs_case('I',EXT2_INDEX_FL)
-            easy_e2fsattrs_case('t',EXT2_NOTAIL_FL)
-            easy_e2fsattrs_case('T',EXT2_TOPDIR_FL)
-#ifdef EXT4_EXTENTS_FL
-            easy_e2fsattrs_case('e',EXT4_EXTENTS_FL)
-#endif
-#ifdef EXT4_HUGE_FILE_FL
-            easy_e2fsattrs_case('h',EXT4_HUGE_FILE_FL)
-#endif
-#ifdef FS_NOCOW_FL
-            easy_e2fsattrs_case('C',FS_NOCOW_FL)
-#endif
-#ifdef EXT4_INLINE_DATA_FL
-            easy_e2fsattrs_case('N',EXT4_INLINE_DATA_FL)
-#endif
-            case '0': {
-                 break;
-            }
-            default: {
-                 LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, ignore invalid ext2 file attribute: '%c', *val)
-                 break;
+    if (strcmp(val, "0") != 0) {
+        while (*val) {
+            unsigned long flag = e2fsattrs_get_flag(*val);
+            if (flag) {
+                conf->report_ignore_e2fsattrs |= flag;
+            } else {
+                 LOG_CONFIG_FORMAT_LINE(LOG_LEVEL_NOTICE, "ignore invalid ext2 file attribute: '%c'", *val)
             }
+            val++;
         }
-        val++;
     }
 }
 #endif
